import math

import pytest

import autofit as af
from test_autofit import mock


class TestPriorLimits:
    def test_out_of_order_prior_limits(self):
        with pytest.raises(af.exc.PriorException):
            af.UniformPrior(1.0, 0)
        with pytest.raises(af.exc.PriorException):
            af.GaussianPrior(0, 1, 1, 0)

    def test_in_or_out(self):
        prior = af.GaussianPrior(0, 1, 0, 1)
        with pytest.raises(af.exc.PriorLimitException):
            prior.assert_within_limits(-1)

        with pytest.raises(af.exc.PriorLimitException):
            prior.assert_within_limits(1.1)

        prior.assert_within_limits(0.0)
        prior.assert_within_limits(0.5)
        prior.assert_within_limits(1.0)

    def test_no_limits(self):
        prior = af.GaussianPrior(0, 1)

        prior.assert_within_limits(100)
        prior.assert_within_limits(-100)
        prior.assert_within_limits(0)
        prior.assert_within_limits(0.5)

    def test_uniform_prior(self):
        prior = af.UniformPrior(0, 1)

        with pytest.raises(af.exc.PriorLimitException):
            prior.assert_within_limits(-1)

        with pytest.raises(af.exc.PriorLimitException):
            prior.assert_within_limits(1.1)

        prior.assert_within_limits(0.0)
        prior.assert_within_limits(0.5)
        prior.assert_within_limits(1.0)

    def test_prior_creation(self):

        mapper = af.ModelMapper()
        mapper.component = mock.MockClassx2

        prior_tuples = mapper.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.lower_limit == 0
        assert prior_tuples[0].prior.upper_limit == 1

        assert prior_tuples[1].prior.lower_limit == 0
        assert prior_tuples[1].prior.upper_limit == 2

    def test_out_of_limits(self):
        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        assert mm.instance_from_vector([1, 2]) is not None

        with pytest.raises(af.exc.PriorLimitException):
            mm.instance_from_vector(([1, 3]))

        with pytest.raises(af.exc.PriorLimitException):
            mm.instance_from_vector(([-1, 2]))

    def test_inf(self):
        mm = af.ModelMapper()
        mm.mock_class_inf = mock.MockClassInf

        prior_tuples = mm.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.lower_limit == -math.inf
        assert prior_tuples[0].prior.upper_limit == 0

        assert prior_tuples[1].prior.lower_limit == 0
        assert prior_tuples[1].prior.upper_limit == math.inf

        assert mm.instance_from_vector([-10000, 10000]) is not None

        with pytest.raises(af.exc.PriorLimitException):
            mm.instance_from_vector(([1, 0]))

        with pytest.raises(af.exc.PriorLimitException):
            mm.instance_from_vector(([0, -1]))

    def test_preserve_limits_tuples(self):
        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        new_mapper = mm.mapper_from_gaussian_tuples(tuples=[(0.0, 0.5), (0.0, 1)], use_widths=True, use_errors=True)

        prior_tuples = new_mapper.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.lower_limit == 0
        assert prior_tuples[0].prior.upper_limit == 1

        assert prior_tuples[1].prior.lower_limit == 0
        assert prior_tuples[1].prior.upper_limit == 2

    def test__only_use_passed_errors_to_set_up_gaussian_prior(self):
        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        new_mapper = mm.mapper_from_gaussian_tuples(
            tuples=[(0.1, 0.2), (0.3, 0.4)],
            use_widths=False,
            use_errors=True
        )

        prior_tuples = new_mapper.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.mean == 0.1
        assert prior_tuples[0].prior.sigma == 0.2

        assert prior_tuples[1].prior.mean == 0.3
        assert prior_tuples[1].prior.sigma == 0.4

    def test__only_use_widths_to_pass_priors(self):

        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        new_mapper = mm.mapper_from_gaussian_tuples(
            tuples=[(5.0, 5.0), (5.0, 5.0)],
            use_widths=True,
            use_errors=False
        )

        prior_tuples = new_mapper.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.mean == 5.0
        assert prior_tuples[0].prior.sigma == 1.0

        assert prior_tuples[1].prior.mean == 5.0
        assert prior_tuples[1].prior.sigma == 2.0

    def test__use_max_of_widths_and_passed_errors_to_pass_priors(self):

        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        new_mapper = mm.mapper_from_gaussian_tuples(
            tuples=[(5.0, 0.2), (5.0, 5.0)],
            use_widths=True,
            use_errors=True
        )

        prior_tuples = new_mapper.prior_tuples_ordered_by_id

        assert prior_tuples[0].prior.mean == 5.0
        assert prior_tuples[0].prior.sigma == 1.0

        assert prior_tuples[1].prior.mean == 5.0
        assert prior_tuples[1].prior.sigma == 5.0

    def test_from_gaussian_no_limits(self):
        mm = af.ModelMapper()
        mm.mock_class_gaussian = mock.MockClassx2

        new_mapper = mm.mapper_from_gaussian_tuples(
            [(0.0, 0.5), (0.0, 1)],
            no_limits=True
        )

        priors = new_mapper.priors
        assert priors[0].lower_limit == float("-inf")
        assert priors[0].upper_limit == float("inf")
        assert priors[1].lower_limit == float("-inf")
        assert priors[1].upper_limit == float("inf")


class TestPriorMean:
    def test_simple(self):
        uniform_prior = af.UniformPrior(0.0, 1.0)
        assert uniform_prior.mean == 0.5

        uniform_prior.mean = 1.0
        assert uniform_prior.lower_limit == 0.5
        assert uniform_prior.upper_limit == 1.5

    def test_higher(self):
        uniform_prior = af.UniformPrior(1.0, 2.0)
        assert uniform_prior.mean == 1.5

        uniform_prior.mean = 2.0
        assert uniform_prior.lower_limit == 1.5
        assert uniform_prior.upper_limit == 2.5


class TestAddition:
    def test_abstract_plus_abstract(self):
        one = af.AbstractModel()
        two = af.AbstractModel()
        one.a = "a"
        two.b = "b"

        three = one + two

        assert three.a == "a"
        assert three.b == "b"

    def test_list_properties(self):
        one = af.AbstractModel()
        two = af.AbstractModel()
        one.a = ["a"]
        two.a = ["b"]

        three = one + two

        assert three.a == ["a", "b"]

    def test_instance_plus_instance(self):
        one = af.ModelInstance()
        two = af.ModelInstance()
        one.a = "a"
        two.b = "b"

        three = one + two

        assert three.a == "a"
        assert three.b == "b"

    def test_mapper_plus_mapper(self):
        one = af.ModelMapper()
        two = af.ModelMapper()
        one.a = af.PriorModel(mock.MockClassx2)
        two.b = af.PriorModel(mock.MockClassx2)

        three = one + two

        assert three.prior_count == 4


class TestUniformPrior:
    def test__simple_assumptions(self):
        uniform_simple = af.UniformPrior(lower_limit=0.0, upper_limit=1.0)

        assert uniform_simple.value_for(0.0) == 0.0
        assert uniform_simple.value_for(1.0) == 1.0
        assert uniform_simple.value_for(0.5) == 0.5

    def test__non_zero_lower_limit(self):
        uniform_half = af.UniformPrior(lower_limit=0.5, upper_limit=1.0)

        assert uniform_half.value_for(0.0) == 0.5
        assert uniform_half.value_for(1.0) == 1.0
        assert uniform_half.value_for(0.5) == 0.75

    def test_width(self):
        assert af.UniformPrior(2, 5).width == 3

    def test_negative_range(self):
        prior = af.UniformPrior(-1, 0)
        assert prior.width == 1
        assert prior.value_for(0.0) == -1
        assert prior.value_for(1.0) == 0.0

    def test__log_prior_from_value(self):

        gaussian_simple = af.UniformPrior(lower_limit=-40, upper_limit=70)

        log_prior = gaussian_simple.log_prior_from_value(value=0.0)

        assert log_prior == 0.0

        log_prior = gaussian_simple.log_prior_from_value(value=11.0)

        assert log_prior == 0.0


class TestLogUniformPrior:
    def test__simple_assumptions(self):
        log_uniform_simple = af.LogUniformPrior(lower_limit=1.0e-8, upper_limit=1.0)

        assert log_uniform_simple.value_for(0.0) == 1.0e-8
        assert log_uniform_simple.value_for(1.0) == 1.0
        assert log_uniform_simple.value_for(0.5) == 0.0001

    def test__non_zero_lower_limit(self):
        log_uniform_half = af.LogUniformPrior(lower_limit=0.5, upper_limit=1.0)

        assert log_uniform_half.value_for(0.0) == 0.5
        assert log_uniform_half.value_for(1.0) == 1.0
        assert log_uniform_half.value_for(0.5) == pytest.approx(0.70710678118, 1.0e-4)

    def test__log_prior_from_value(self):

        gaussian_simple = af.LogUniformPrior(lower_limit=0.0, upper_limit=1.0)

        log_prior = gaussian_simple.log_prior_from_value(value=1.0)

        assert log_prior == 1.0

        log_prior = gaussian_simple.log_prior_from_value(value=2.0)

        assert log_prior == 0.5

        log_prior = gaussian_simple.log_prior_from_value(value=4.0)

        assert log_prior == 0.25

        gaussian_simple = af.LogUniformPrior(lower_limit=50.0, upper_limit=100.0)

        log_prior = gaussian_simple.log_prior_from_value(value=1.0)

        assert log_prior == 1.0

        log_prior = gaussian_simple.log_prior_from_value(value=2.0)

        assert log_prior == 0.5

        log_prior = gaussian_simple.log_prior_from_value(value=4.0)

        assert log_prior == 0.25


class TestGaussianPrior:
    def test__simple_assumptions(self):
        gaussian_simple = af.GaussianPrior(mean=0.0, sigma=1.0)

        assert gaussian_simple.value_for(0.1) == pytest.approx(-1.281551, 1.0e-4)
        assert gaussian_simple.value_for(0.9) == pytest.approx(1.281551, 1.0e-4)
        assert gaussian_simple.value_for(0.5) == 0.0

    def test__non_zero_mean(self):
        gaussian_half = af.GaussianPrior(mean=0.5, sigma=2.0)

        assert gaussian_half.value_for(0.1) == pytest.approx(-2.0631031, 1.0e-4)
        assert gaussian_half.value_for(0.9) == pytest.approx(3.0631031, 1.0e-4)
        assert gaussian_half.value_for(0.5) == 0.5

    def test__log_prior_from_value(self):

        gaussian_simple = af.GaussianPrior(mean=0.0, sigma=1.0)

        log_prior = gaussian_simple.log_prior_from_value(value=0.0)

        assert log_prior == 0.0

        log_prior = gaussian_simple.log_prior_from_value(value=1.0)

        assert log_prior == 0.5

        log_prior = gaussian_simple.log_prior_from_value(value=2.0)

        assert log_prior == 2.0

        gaussian_simple = af.GaussianPrior(mean=1.0, sigma=2.0)

        log_prior = gaussian_simple.log_prior_from_value(value=0.0)

        assert log_prior == 0.125

        log_prior = gaussian_simple.log_prior_from_value(value=1.0)

        assert log_prior == 0.0

        log_prior = gaussian_simple.log_prior_from_value(value=2.0)

        assert log_prior == 0.125

        gaussian_simple = af.GaussianPrior(mean=30.0, sigma=60.0)

        log_prior = gaussian_simple.log_prior_from_value(value=2.0)

        assert log_prior == pytest.approx(0.108888, 1.0e-4)


# %%
"""
Tutorial 7: Phase Customization
===============================

In this tutorial, we're going to add input parameters to a Phase object that customizes the analysis. We'll use the
specific example of two input parameters that trim our data-set from the left and right before fitting it. This
example is somewhat trivial (we could achieve almost the same effect with masking), but it will serve to illustrate
phase customization.

When we customize a phase, we'll use these 'phase settings' to perform phase tagging. Here, we 'tag' the output
path of the phase's results, such that every time a phase is run with a different customization a new set of
unique results are stored for those settings. For a given data-set we are thus able to fit it multiple times using
different settings to compare the results.

These new features have led to additional modules in the 'phase' package called 'meta_dataset.py' and 'settings.py'.
Before looking at these modules, lets first perform a series of Emcee fits to see how they change the behaviour
of PyAutoFit.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
import numpy as np

from howtofit.chapter_1_introduction.tutorial_7_phase_customization import src as htf

from pyprojroot import here

workspace_path = str(here())
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our non-linear search.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/howtofit/config",
    output_path=f"{workspace_path}/howtofit/output",
)

# %%
"""
We're now going to perform multiple fits, where each fit trims the data-set that is fitted.
To do this, we'll set up phases with the phase-settings 'data_trim_left' and 'data_trim_right'.

- data_trim_left:

  The dataset's image and noise-map are trimmed and removed from the left (e.g. 1d index values from 0).
  For example, if the dataset has shape (100,) and we set data_trim_left=10, the dataset that is fitted will have
  shape (90,). The mask is trimmed in the same way.

- data_trim_right:

  This behaves the same as data_trim_left, however data is removed from the right (e.g. 1D index values from the
  shape of the 1D data).

For our first phase, we will omit both the phase setting (by setting it to None) and perform the fit from tutorial
4 where we fit a single Gaussian profile to data composed of a single Gaussian (unlike tutorial 4, we'll use a
CollectionPriorModel to do this).
"""

# %%
phase = htf.Phase(
    phase_name="phase_t7",
    profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
    settings=htf.SettingsPhase(data_trim_left=None, data_trim_right=None),
    search=af.Emcee(),
)

# %%
"""
Import the simulator module, set up the Dataset and mask and set up the dataset.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

dataset = htf.Dataset(data=gaussian_x1.data, noise_map=gaussian_x1.noise_map)
mask = np.full(fill_value=False, shape=dataset.data.shape)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t6"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Okay, lets look at what happened differently in this phase. To begin, lets note the output directory:

'autofit_workspace/howtofit/chapter_1_introduction/tutorial_7_phase_customization/output/phase_t7/settings'

There is a small change to this directory compared to tutorial 6, there is a new folder 'settings' within which the
results are stored. It'll be clear why this is in a moment.

Next, we're going to customize and run a phase using the data_trim_left and right parameters. First, we create a 
SettingsPhase object using our input values of these parameters. 
"""

# %%
settings = htf.SettingsPhase(data_trim_left=20, data_trim_right=30)

# %%
"""
We now create a new phase with these settings and run it (note that we haven't changed the phase_name from 'phase_t7', 
which you might think would cause conflicts in the path the results are output to).
"""

# %%
phase = htf.Phase(
    phase_name="phase_t7",
    profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
    settings=settings,
    search=af.Emcee(),
)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t6"
    "folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

phase.run(dataset=dataset, mask=mask)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
You'll note the results are now in a slightly different directory to the fit performed above:

 'autofit_workspace/howtofit/chapter_1_introduction/tutorial_7_phase_customization/output/phase_example/settings__trim_left_20__trim_right_30'

By customizing the phase's settings, PyAutoFit has changed it output path using a tag for this phase. There are two
reasons PyAutoFit does this:

 1) Tags describes the analysis, making it explicit what was done to the dataset for the fit.

 2) Tags create a unique output path, allowing you to compare results of phases that use different settings. Equally,
 if you run multiple phases with different settings this ensures the non-linear search (e.g. Emcee) won't
 inadvertantly use results generated via a different analysis method.

You should now check out the 'settings.py' and 'meta_dataset.py' modules in the 'phase' package, to see how we 
implemented this.

In this tutorial, the phase setting changed the data-set that was fitted. However, phase settings do not necessarily
need to customize the data-set. For example, they could control some aspect of the model, for example the precision
by which the model image is numerically calculated. For more complex fitting procedures, settings could control
whether certain features are used, which when turned on / off reduce the accuracy of the model at the expensive of
greater computational run-time.

Phase settings are project specific and it could well be your modeling problem is simple enough not to need them.
However, if it does, remember that phase settings are a powerful means to fit models using different settings and
compare whether a setting does or does not change the model inferred. In later chapters, we'll discuss more complex
model-fitting procedures that could use 'fast' less accurate settings to initialize the model-fit, but switch to
slower more accurate settings later on.
"""



import numpy as np

from autofit.message_passing.messages.abstract import AbstractMessage


class FixedMessage(AbstractMessage):
    log_base_measure = 0

    def __init__(self, value, log_norm=0.):
        self._value = value
        super().__init__(
            (value,),
            log_norm=log_norm
        )

    @property
    def natural_parameters(self):
        return self.parameters

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        return natural_parameters,

    @staticmethod
    def to_canonical_form(x):
        return x

    @property
    def log_partition(self):
        return 0.

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        return suff_stats

    def sample(self, n_samples):
        """
        Rely on array broadcasting to get fixed values to
        calculate correctly
        """
        return np.array(self.parameters)

    def logpdf(self, x):
        return np.zeros_like(x)

    @property
    def mean(self):
        return self._value

    @property
    def variance(self):
        return np.zeros_like(self.mean)

    def _no_op(self, *other, **kwargs):
        """
        'no-op' operation

        In many operations fixed messages should just
        return themselves
        """
        return self

    project = _no_op
    from_mode = _no_op
    __pow__ = _no_op
    __mul__ = _no_op
    __div__ = _no_op
    default = _no_op
    _multiply = _no_op
    _divide = _no_op
    sum_natural_parameters = _no_op
    sub_natural_parameters = _no_op



import numpy as np

from autofit.message_passing.messages.abstract import AbstractMessage


class NormalMessage(AbstractMessage):
    @property
    def log_partition(self):
        eta1, eta2 = self.natural_parameters
        return - eta1 ** 2 / 4 / eta2 - np.log(-2 * eta2) / 2

    log_base_measure = - 0.5 * np.log(2 * np.pi)
    _support = ((-np.inf, np.inf),)
    _parameter_support = ((-np.inf, np.inf), (0, np.inf))

    def __init__(
            self,
            mu=0.,
            sigma=1.,
            log_norm=0.
    ):
        self.mu = mu
        self.sigma = sigma
        super().__init__(
            (mu, sigma),
            log_norm=log_norm
        )

    @classmethod
    def from_prior(
            cls,
            prior
    ):
        return NormalMessage(
            mu=prior.mean,
            sigma=prior.sigma
        )

    @property
    def natural_parameters(self):
        return self.calc_natural_parameters(
            self.mu,
            self.sigma
        )

    @staticmethod
    def calc_natural_parameters(mu, sigma):
        precision = sigma ** -2
        return np.array([mu * precision, - precision / 2])

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        eta1, eta2 = natural_parameters
        mu = - 0.5 * eta1 / eta2
        sigma = np.sqrt(- 0.5 / eta2)
        return mu, sigma

    @staticmethod
    def to_canonical_form(x):
        return np.array([x, x ** 2])

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        m1, m2 = suff_stats
        sigma = np.sqrt(m2 - m1 ** 2)
        return cls.calc_natural_parameters(m1, sigma)

    @property
    def mean(self):
        return self.mu

    @property
    def variance(self):
        return self.sigma ** 2

    def sample(self, n_samples):
        x = np.random.randn(n_samples, *self.shape)
        mu, sigma = self.parameters
        if self.shape:
            return x * sigma[None, ...] + mu[None, ...]

        return x * sigma + mu

    @classmethod
    def from_mode(cls, mode: np.ndarray, covariance: float = 1.):
        mode, variance = cls._get_mean_variance(mode, covariance)
        return cls(mode, variance ** 0.5)
